package de.tudresden.inf.st.ros2rag.example;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.fusesource.hawtbuf.Buffer;
import org.fusesource.hawtbuf.UTF8Buffer;
import org.fusesource.mqtt.client.*;

import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Consumer;

/**
 * Helper class to receive updates via MQTT and use callbacks to handle those messages.
 *
 * @author rschoene - Initial contribution
 */
public class MqttUpdater {

  private final Logger logger;

  /** The host running the MQTT broker. */
  private URI host;
  /** The connection to the MQTT broker. */
  private CallbackConnection connection;
  /** Whether we are subscribed to the topics yet */
  private Condition readyCondition;
  private Lock readyLock;
  private boolean ready;
  private QoS qos;
  /** Dispatch knowledge */
  private final Map<String, Consumer<byte[]>> callbacks;

  public MqttUpdater() {
    this.logger = LogManager.getLogger(MqttUpdater.class);
    this.callbacks = new HashMap<>();
    this.readyLock = new ReentrantLock();
    this.readyCondition = readyLock.newCondition();
    this.ready = false;
    this.qos = QoS.AT_LEAST_ONCE;
  }

  /**
   * Sets the host to receive messages from, and connects to it.
   * @throws IOException if could not connect, or could not subscribe to a topic
   * @return self
   */
  public MqttUpdater setHost(String host, int port) throws IOException {
    this.host = URI.create("tcp://" + host + ":" + port);
    logger.debug("Host is {}", this.host);

    Objects.requireNonNull(this.host, "Host need to be set!");
    MQTT mqtt = new MQTT();
    mqtt.setHost(this.host);
    connection = mqtt.callbackConnection();
    AtomicReference<Throwable> error = new AtomicReference<>();

    // add the listener to dispatch messages later
    connection.listener(new ExtendedListener() {
      public void onConnected() {
        logger.debug("Connected");
      }

      @Override
      public void onDisconnected() {
        logger.debug("Disconnected");
      }

      @Override
      public void onPublish(UTF8Buffer topic, Buffer body, Callback<Callback<Void>> ack) {
        String topicString = topic.toString();
        Consumer<byte[]> callback = callbacks.get(topicString);
        if (callback == null) {
          logger.debug("Got a message, but no callback to call. Forgot to unsubscribe?");
        } else {
          byte[] message = body.toByteArray();
//          System.out.println("message = " + Arrays.toString(message));
          callback.accept(message);
        }
        ack.onSuccess(null);  // always acknowledge message
      }

      @Override
      public void onPublish(UTF8Buffer topicBuffer, Buffer body, Runnable ack) {
        logger.warn("onPublish should not be called");
      }

      @Override
      public void onFailure(Throwable cause) {
//        logger.catching(cause);
        error.set(cause);
      }
    });
    throwIf(error);

    // actually establish the connection
    connection.connect(new Callback<Void>() {
      @Override
      public void onSuccess(Void value) {
        connection.publish("components", "Ros2Rag is listening".getBytes(), QoS.AT_LEAST_ONCE, false, new Callback<Void>() {
          @Override
          public void onSuccess(Void value) {
            logger.debug("success sending welcome message");
            try {
              readyLock.lock();
              ready = true;
              readyCondition.signalAll();
            } finally {
              readyLock.unlock();
            }
          }

          @Override
          public void onFailure(Throwable value) {
            logger.debug("failure sending welcome message", value);
          }
        });
      }

      @Override
      public void onFailure(Throwable cause) {
//        logger.error("Could not connect", cause);
        error.set(cause);
      }
    });
    throwIf(error);
    return this;
  }

  private void throwIf(AtomicReference<Throwable> error) throws IOException {
    if (error.get() != null) {
      throw new IOException(error.get());
    }
  }

  public void setQoSForSubscription(QoS qos) {
    this.qos = qos;
  }

  public void newConnection(String topic, Consumer<byte[]> callback) {
    if (!ready) {
      // TODO should maybe be something more kind than throwing an exception here
      throw new IllegalStateException("Updater not ready");
    }
    // register callback
    callbacks.put(topic, callback);

    // subscribe at broker
    Topic[] topicArray = { new Topic(topic, this.qos) };
    connection.subscribe(topicArray, new Callback<byte[]>() {
      @Override
      public void onSuccess(byte[] qoses) {
        logger.debug("Subscribed, qoses: {}", qoses);
      }

      @Override
      public void onFailure(Throwable cause) {
        logger.error("Could not subscribe", cause);
      }
    });
  }

  /**
   * Waits until this updater is ready to receive MQTT messages.
   * If it already is ready, return immediately with the value <code>true</code>.
   * Otherwise waits for the given amount of time, and either return <code>true</code> within the timespan,
   * if it got ready, or <code>false</code> upon a timeout.
   * @param time the maximum time to wait
   * @param unit the time unit of the time argument
   * @return whether this updater is ready
   */
  public boolean waitUntilReady(long time, TimeUnit unit) {
    try {
      readyLock.lock();
      if (ready) {
        return true;
      }
      return readyCondition.await(time, unit);
    } catch (InterruptedException e) {
      e.printStackTrace();
    } finally {
      readyLock.unlock();
    }
    return false;
  }


  public void close() {
    if (connection == null) {
      logger.warn("Stopping without connection. Was setHost() called?");
      return;
    }
    connection.disconnect(new Callback<Void>() {
      @Override
      public void onSuccess(Void value) {
        logger.info("Disconnected from {}", host);
      }

      @Override
      public void onFailure(Throwable ignored) {
        // Disconnects never fail. And we do not care either.
      }
    });
  }

}
